/***************************************************************************
 *   Copyright (C) 2010 by Martin Burgemeister, Mathis Lindemann           *
 *                                                                         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "UCQConsole.h"

/**
 * \brief Constructs the QTextEdit and calls init() (see Luna.h)
 *
 * \param parent Pointer to the parent QWidget.
 * \param prompt String for prompt, e.g. ">".
 * \param history History to use.
 */
UCQConsole::UCQConsole ( QWidget* parent, const QString& prompt,
                         const QStringList& history ) :
    QTextEdit ( parent ), prompt_ ( prompt ), history_ ( history ),
            completeRe_ ( "[_.:a-zA-Z0-9]+$" )
{
    init ();
}

/**
 * \brief Emtpy destructor.
 */
UCQConsole::~UCQConsole ()
{
//    emit stopExecution ();
//    executeLuaThread_->exit (0);
}

/**
 * \brief Handler for End-key events.
 *
 * \param select If true, the text is selected.
 */
void UCQConsole::actionEnd ( bool select )
{
    tc_ = textCursor ();
    tc_.movePosition ( QTextCursor::End, select ? QTextCursor::KeepAnchor
                                                : QTextCursor::MoveAnchor, 1 );
    setTextCursor ( tc_ );
}

/**
 * \brief Shows the next command in history.
 */
void UCQConsole::actionHistoryDown ()
{
    // end of history
    if ( historyCount_ + 1 >= history_.size () ) {
        return;
    }

    setUndoRedoEnabled ( false );
    tc_ = textCursor ();
    // select current text
    tc_.setPosition ( lineStart_, QTextCursor::MoveAnchor );
    tc_.movePosition ( QTextCursor::End, QTextCursor::KeepAnchor, 1 );

    // replace last history entry with selected text
    history_.replace ( historyCount_, tc_.selectedText () );

    // insert text and increment historyCount for next use
    tc_.insertText ( history_[ ++historyCount_ ] );
    setUndoRedoEnabled ( true );
}

/**
 * \brief Search function for history.
 *
 * \param direction positive = forwards, negative = backwards.
 */
void UCQConsole::actionHistoryFind ( int direction )
{
    setUndoRedoEnabled ( false );
    tc_ = textCursor ();
    QString str;
    bool hs;
    // get "searchtext"
    hs = tc_.hasSelection ();
    if ( hs ) {
        str = tc_.selectedText ();
    }

    // select current line
    tc_.setPosition ( lineStart_, QTextCursor::MoveAnchor );
    tc_.movePosition ( QTextCursor::End, QTextCursor::KeepAnchor, 1 );

    // use all input if nothing was selected
    if ( !hs ) {
        str = tc_.selectedText ().trimmed ();
    }

    // find text in history
    if ( str.size () > 0 ) {
        // check every entry in the given direction
        for ( int i = historyCount_ + direction; i >= 0 && i < history_.size (); i
                += direction ) {
            // compare entry to text
            int count = history_[ i ].indexOf ( str );

            // match - insert next/previous matching entry
            if ( count >= 0 ) {
                // set historyCount for the next call
                historyCount_ = i;
                tc_.insertText ( history_[ i ] );
                tc_.setPosition ( lineStart_ + count, QTextCursor::MoveAnchor );
                tc_.setPosition ( lineStart_ + count + str.size (),
                                  QTextCursor::KeepAnchor );
                setTextCursor ( tc_ );
                break;
            }
        }
    }
    setUndoRedoEnabled ( true );
}

/**
 * \brief Shows the previous command in history.
 */
void UCQConsole::actionHistoryUp ()
{
    if ( historyCount_ == 0 ) {
        return;
    }

    setUndoRedoEnabled ( false );
    tc_ = textCursor ();
    // select text
    tc_.setPosition ( lineStart_, QTextCursor::MoveAnchor );
    tc_.movePosition ( QTextCursor::End, QTextCursor::KeepAnchor, 1 );

    // replace last history entry with selected text
    history_.replace ( historyCount_, tc_.selectedText () );

    // insert text and decrement historyCount for next use
    tc_.insertText ( history_[ --historyCount_ ] );
    setUndoRedoEnabled ( true );
}

/**
 * \brief Handler for Home/Pos1-key events.
 *
 * \param select If true, the text is selected.
 */
void UCQConsole::actionHome ( bool select )
{
    tc_ = textCursor ();
    // if select, also select text
    tc_.setPosition ( lineStart_, select ? QTextCursor::KeepAnchor
                                         : QTextCursor::MoveAnchor );
    setTextCursor ( tc_ );
}

/**
 * \brief Autocomplete method.
 *
 * The autocomplete retrieves all functionnames from the LuaController, that
 * start with the text already entered by the user. If multiple hits where
 * found, it displays them above the current line.
 */
void UCQConsole::actionKeyComplete ()
{
    QStringList list;
    // get text chunk to complete
    tc_ = textCursor ();
    tc_.setPosition ( lineStart_, QTextCursor::KeepAnchor );
    QString prefix = tc_.selectedText ();
    int idx = prefix.indexOf ( completeRe_ );

    if ( idx < 0 ) {
        deleteCompletionList ();
        return;
    }

    int offset = 0;
    // get completion candidates list
    emit getCompletionList ( prefix.mid ( idx ), list );

    switch ( list.count () ) {
        // no candidate
        case 0:
            deleteCompletionList ();
            return;

        // single candidate
        case 1: {
            tc_ = textCursor ();

            // insert
            tc_.setPosition ( lineStart_ + idx, QTextCursor::KeepAnchor );
            tc_.insertText ( list[ 0 ] );

            // move cursor if needed
            if ( offset ) {
                tc_.setPosition ( tc_.position () + offset,
                                  QTextCursor::MoveAnchor );
                setTextCursor ( tc_ );
            }
            deleteCompletionList ();
            break;
        }

        // multiple candidates
        default: {
            int i;
            QString str;
            // find common prefix
            for ( i = 0;; i++ ) {
                QChar c;
                bool end = false;
                for ( int j = 0; j < list.size (); ++j ) {
                    str = list.at ( j );
                    if ( i == str.size () ) {
                        end = true;
                        break;
                    }

                    if ( c.isNull () ) {
                        c = str[ i ];
                    } else if ( str[ i ] != c ) {
                        end = true;
                        break;
                    }
                }
                if ( end ) {
                    break;
                }
            }

            // insert common prefix
            if ( i > 0 ) {
                tc_ = textCursor ();

                tc_.setPosition ( lineStart_ + idx, QTextCursor::KeepAnchor );
                tc_.insertText ( list[ 0 ].left ( i ) );
            }

            // clear previous completion list text
            tc_ = textCursor ();
            int offset = tc_.position () - lineStart_;
            tc_.setPosition ( completeStart_, QTextCursor::MoveAnchor );
            tc_.setPosition ( promptStart_, QTextCursor::KeepAnchor );
            tc_.removeSelectedText ();
            setTextCursor ( tc_ );

            // insert new completion list text
            for ( int i = 0; i < list.size (); ++i ) {
                str = list.at ( i );
                setTextColor ( Qt::darkRed );
                QString s = str;
                s = s.leftJustified ( 60, ' ' );
                insertPlainText ( s );
            }

            insertPlainText ( "\n" );
            setTextColor ( Qt::black );

            // adjust cursor position variables
            tc_ = textCursor ();
            lineStart_ += tc_.position () - promptStart_;
            mark_ += tc_.position () - promptStart_;
            promptStart_ = tc_.position ();
            tc_.setPosition ( lineStart_ + offset, QTextCursor::MoveAnchor );
            setTextCursor ( tc_ );
        }
    }
}

/**
 * \brief Method for handling CR-key events.
 *
 * This method takes the string from the last line and passed it to the
 * LuaController for execution. It also updates the history and displays an
 * error if the execution failed. If the last line is empty, nothing happens.
 *
 * \see executionInit()
 * \see executionReady()
 */
void UCQConsole::actionKeyEnter ()
{
    tc_ = textCursor ();
    tc_.movePosition ( QTextCursor::End, QTextCursor::MoveAnchor, 1 );
    setTextCursor ( tc_ );
    // Get the line
    tc_.setPosition ( lineStart_, QTextCursor::MoveAnchor );
    tc_.movePosition ( QTextCursor::End, QTextCursor::KeepAnchor, 1 );

    QString line = tc_.selectedText ().trimmed ();

    if ( line.isEmpty () ) {
        append ( "" );
        displayPrompt();
        return;
    }

    // Checks history size
    while ( history_.size () > historySize_ ) {
        history_.removeFirst ();
    }

    // Detect duplicate entries in history
    if ( history_.size () > 1 ) {
        int prev = history_.lastIndexOf ( line, history_.size () - 2 );

        if ( prev >= 0 ) {
            history_.removeAt ( prev );
        }
    }

    // Set the history entry
    history_.replace ( history_.size () - 1, line );
    history_.append ( "" );

    historyCount_ = history_.size () - 1;

    // Move visible cursor to end
    tc_.clearSelection ();
    setTextCursor ( tc_ );

    // append single words "commands" with ()
    bool singleCommand = true;
    for ( int i = 0; i < line.size (); ++i ) {
        if ( !line.at ( i ).isLetterOrNumber () ) {
            singleCommand = false;
        }
    }
    if ( singleCommand ) {
        line.append ( "()" );
    }

    // check for help flag
    if ( line.indexOf ( "-h", 0 ) > -1 ) {
        line.insert ( line.indexOf ( "-h" ), "\"" );
        line.insert ( line.indexOf ( "-h" ) + 2, "\"" );
    }
    executeLuaThread_->setexecutionMode(UCQExecuteLuaThread::EXECUTE_STRING);
    executeLuaThread_->execute ( line );



}

/**
 * \brief Checks if the textcursor is in the last line and may accept input.
 */
void UCQConsole::checkWritingPosition ()
{
    tc_ = textCursor ();
    if ( tc_.hasSelection () ) {

        // invalid selection
        if ( tc_.selectionStart () < lineStart_ ) {
            tc_.clearSelection ();
            setTextCursor ( tc_ );
        }
    }

    // move cursor to last line if necessary
    int pos = tc_.position ();

    tc_.movePosition ( QTextCursor::End, QTextCursor::MoveAnchor, 1 );

    if ( pos < lineStart_ ) {
        setTextCursor ( tc_ );
    }
}

/**
 * \brief Clears the console and sets a new prompt.
 */
void UCQConsole::clearConsole ()
{
    clear ();
    displayPrompt ();
}

/**
 * \brief Removes the text printed by the actionKeyComplete() method from the
 *        console.
 */
void UCQConsole::deleteCompletionList ()
{
    if ( completeStart_ != promptStart_ ) {
        tc_ = textCursor ();

        // remove text from first column till next prompt
        tc_.setPosition ( completeStart_, QTextCursor::MoveAnchor );
        tc_.setPosition ( promptStart_, QTextCursor::KeepAnchor );
        tc_.removeSelectedText ();

        // update positions
        lineStart_ += tc_.position () - promptStart_;
        mark_ += tc_.position () - promptStart_;
        promptStart_ = tc_.position ();
    }
}

/**
 * \brief Display a new prompt.
 */
void UCQConsole::displayPrompt ()
{
    setUndoRedoEnabled ( false );
    tc_ = textCursor ();
    completeStart_ = promptStart_ = tc_.position ();

    setTextColor ( Qt::green );
    insertPlainText ( prompt_ );

    setTextColor ( Qt::black );
    tc_ = textCursor ();
    insertPlainText ( " " );

    mark_ = lineStart_ = tc_.position ();

    setUndoRedoEnabled ( true );
}

/**
 * \brief Displays a (error-)message in the console.
 *
 * \param message The message to display.
 */
void UCQConsole::error ( QString message )
{
    setTextColor ( Qt::red );
    append ( message );
}

/**
 * \brief Put the console in the execution state.
 *
 * Sets the console read only and starts the timer for showing busy dots.
 * \see executionRuns()
 */
void UCQConsole::executionInit ()
{
    tc_ = textCursor ();

    // get position for busy dots
    executionStart_ = tc_.position ();
    executionEnd_ = executionStart_ + 1;
    setReadOnly ( true );
    insertPlainText ( "\n" );
    setTextInteractionFlags ( Qt::NoTextInteraction );

    // start displaying busy dots
    timer->start ( 1000 );
    executing_ = true;
}

/**
 * \brief Puts the console back in normal state when execution is finished.
 *
 * Stops the busy dots and makes the console writable again.
 * \see executionRuns()
 * \see executionInit()
 */
void UCQConsole::executionReady ()
{
    executing_ = false;

    // stop busy dots
    timer->stop ();
    tc_ = textCursor ();

    // remove dots
    tc_.setPosition ( executionStart_, QTextCursor::MoveAnchor );
    tc_.setPosition ( executionEnd_, QTextCursor::KeepAnchor );
    tc_.removeSelectedText ();

    // display new prompt
    tc_.movePosition ( QTextCursor::End, QTextCursor::MoveAnchor, 1 );
    tc_.insertText ( "\n" );
    setTextCursor ( tc_ );
    displayPrompt ();
    setReadOnly ( false );
    setTextInteractionFlags ( Qt::TextEditorInteraction );
}

/**
 * \brief Execution state of the console.
 *
 * Shows busy dots while the console is executing a long running command.
 *
 * \see executionInit()
 */
void UCQConsole::executionRuns ()
{
    tc_ = textCursor ();

    // move to position for busy dots
    tc_.setPosition ( executionStart_ + 1, QTextCursor::MoveAnchor );
    tc_.setPosition ( executionEnd_, QTextCursor::KeepAnchor );
    QString dots = tc_.selectedText ();

    // three dots max
    if ( dots == "..." ) {
        tc_.removeSelectedText ();
        executionEnd_ = executionStart_ + 1;
    }

    // insert one dot
    tc_.setPosition ( executionEnd_, QTextCursor::MoveAnchor );
    tc_.insertText ( "." );
    tc_.clearSelection ();
    tc_.movePosition ( QTextCursor::End, QTextCursor::MoveAnchor );
    setTextCursor ( tc_ );

    executionEnd_++;
}

/**
 * \brief Gets the prompt.
 *
 * \return String that represents a Prompt.
 */
const QString& UCQConsole::getPrompt () const
{
    return prompt_;
}

/**
 * \brief Initializes the console.
 */
void UCQConsole::init ()
{
    timer = new QTimer ( this );
    executing_ = false;
    connect ( timer, SIGNAL( timeout() ),
              this,  SLOT  ( executionRuns() ) );

    mark_           = 0;
    lineStart_      = 0;
    cursorPos_      = 0;
    promptStart_    = 0;
    executionEnd_   = 0;
    completeStart_  = 0;
    executionStart_ = 0;

    formatNormal_.setFontFamily ( "Monofont" );
    formatNormal_.setFontFixedPitch ( true );
    formatNormal_.setFontPointSize ( 10 );

    setCurrentCharFormat ( formatNormal_ );
    setWordWrapMode ( QTextOption::WrapAnywhere );

    historyCount_ = history_.size ();
    historySize_ = 100;
    history_.append ( "" );
    insertPlainText ( tr ( "Lua Scripting" ) );
}

/**
 * \brief Eventhandler for keyboard.
 *
 * All shortkeys are defined in this method.
 *
 * \param e QKeyEvent.
 */
void UCQConsole::keyPressEvent ( QKeyEvent* e )
{
    // Ctrl+N stops the execution when nothing else works
    if ( executing_ ) {
        if ( e->modifiers () == Qt::ControlModifier && e->key () == Qt::Key_N ) {
            emit stopExecution ();
        }
    } else {

        int cur;
        switch ( e->modifiers () ) {
            case Qt::ControlModifier:
                switch ( e->key () ) {
                    // select whole line
                    case Qt::Key_A:
                        tc_ = textCursor ();
                        tc_.setPosition ( lineStart_, QTextCursor::MoveAnchor );
                        tc_.movePosition ( QTextCursor::End,
                                           QTextCursor::KeepAnchor, 1 );
                        setTextCursor ( tc_ );
                        break;

                    // copy
                    case Qt::Key_C:
                        copy ();
                        break;

                    // forward find
                    case Qt::Key_F:
                        actionHistoryFind ( 1 );
                        break;

                    // Cut line from current position to end
                    case Qt::Key_K:
                        tc_ = textCursor ();
                        tc_.movePosition ( QTextCursor::End,
                                           QTextCursor::KeepAnchor, 1 );
                        setTextCursor ( tc_ );
                        cut ();
                        break;

                    // clear prompt */
                    case Qt::Key_L:
                        if ( !executing_ )
                            clearConsole ();
                        break;

                    // set the mark to the current position
                    case Qt::Key_M:
                        mark_ = textCursor ().position ();
                        break;

                    // new line
                    case Qt::Key_N:

                        tc_ = textCursor ();
                        tc_.movePosition ( QTextCursor::End,
                                           QTextCursor::MoveAnchor, 1 );
                        tc_.insertText ( "\n" );
                        setTextCursor ( tc_ );
                        displayPrompt ();

                        break;

                    // reverse find
                    case Qt::Key_R:
                        actionHistoryFind ( -1 );
                        break;

                    // paste
                    case Qt::Key_V:
                        paste ();
                        break;

                    // cut line from current position to mark
                    case Qt::Key_W:
                        tc_ = textCursor ();
                        tc_.setPosition ( mark_, QTextCursor::KeepAnchor );
                        setTextCursor ( tc_ );
                        cut ();
                        break;

                    // Cut the current line
                    case Qt::Key_X:
                        tc_ = textCursor ();
                        tc_.setPosition ( lineStart_, QTextCursor::KeepAnchor );
                        setTextCursor ( tc_ );
                        cut ();
                        break;

                    // redo
                    case Qt::Key_Y:
                        redo ();
                        break;

                    // undo
                    case Qt::Key_Z:
                        undo ();
                        break;

                    // jump to previous word
                    case Qt::Key_Left:
                        tc_ = textCursor ();
                        tc_.movePosition ( QTextCursor::PreviousWord,
                                           QTextCursor::MoveAnchor, 1 );
                        if ( tc_.position () < lineStart_ ) {
                            tc_.setPosition ( lineStart_,
                                              QTextCursor::MoveAnchor );
                        }
                        setTextCursor ( tc_ );
                        break;

                    // jump to next word
                    case Qt::Key_Right:
                        tc_ = textCursor ();
                        tc_.movePosition ( QTextCursor::NextWord,
                                           QTextCursor::MoveAnchor, 1 );
                        setTextCursor ( tc_ );
                        break;

                    // remove previous word
                    case Qt::Key_Backspace:

                        tc_ = textCursor ();
                        tc_.movePosition ( QTextCursor::PreviousWord,
                                           QTextCursor::KeepAnchor, 1 );
                        if ( tc_.position () < lineStart_ ) {
                            tc_.setPosition ( lineStart_,
                                              QTextCursor::KeepAnchor );
                        }
                        tc_.removeSelectedText ();
                        break;
                }
                ensureCursorVisible ();
                break;

            case Qt::AltModifier:
                switch ( e->key () ) {

                    // remove next word
                    case Qt::Key_D:
                        tc_ = textCursor ();
                        tc_.movePosition ( QTextCursor::NextWord,
                                           QTextCursor::KeepAnchor, 1 );
                        tc_.removeSelectedText ();
                        break;

                    // copy from current position to mark
                    case Qt::Key_W:
                        tc_ = textCursor ();
                        cur = tc_.position ();
                        tc_.setPosition ( mark_, QTextCursor::KeepAnchor );
                        setTextCursor ( tc_ );
                        copy ();
                        tc_.setPosition ( cur, QTextCursor::MoveAnchor );
                        setTextCursor ( tc_ );
                        break;
                }
                ensureCursorVisible ();
                break;

            case Qt::KeypadModifier:
            case Qt::NoModifier:
                // ordinary input ( characters, numbers,...
                if ( e->key () >= Qt::Key_Space && e->key ()
                        <= Qt::Key_AsciiTilde && e->key () != Qt::Key_Comma ) {
                    checkWritingPosition ();
                    QTextEdit::keyPressEvent ( e );
                    ensureCursorVisible ();
                    break;
                }
                switch ( e->key () ) {

                    // jump to end of line
                    case Qt::Key_End:
                        actionEnd ( false );
                        break;

                    // make sure prompt is not deleted
                    case Qt::Key_Delete:
                        if ( textCursor ().selectionStart () >= lineStart_ ) {
                            QTextEdit::keyPressEvent ( e );
                        }
                        break;

                    case Qt::Key_Right:
                        QTextEdit::keyPressEvent ( e );
                        break;

                    // execute command
                    case Qt::Key_Return:
                    case Qt::Key_Enter:
                        actionKeyEnter ();
                        break;

                    // protect prompt
                    case Qt::Key_Left:
                    case Qt::Key_Backspace:
                        if ( textCursor ().position () > lineStart_ ) {
                            QTextEdit::keyPressEvent ( e );
                        }
                        break;

                    // jump to first input column ( right of promt )
                    case Qt::Key_Home:
                        actionHome ( false );
                        break;

                    // previous history entry
                    case Qt::Key_Up:
                        actionHistoryUp ();
                        break;

                    // next history entry
                    case Qt::Key_Down:
                        actionHistoryDown ();
                        break;

                    // one page up
                    case Qt::Key_PageUp:
                        verticalScrollBar ()->triggerAction (
                                QAbstractSlider::SliderPageStepSub );
                        return;

                    // one page down
                    case Qt::Key_PageDown:
                        verticalScrollBar ()->triggerAction (
                                QAbstractSlider::SliderPageStepAdd );
                        return;

                    // autocomplete
                    case Qt::Key_Tab:
                        actionKeyComplete ();
                        break;

                    // inserts "" after a comma for comfort
                    case Qt::Key_Comma:
                        QTextEdit::keyPressEvent ( e );
                        tc_ = textCursor ();
                        tc_.movePosition ( QTextCursor::Right,
                                           QTextCursor::KeepAnchor, 1 );
                        if ( tc_.selectedText () != "\"" ) {
                            tc_.clearSelection ();
                            tc_.movePosition ( QTextCursor::Left,
                                               QTextCursor::MoveAnchor, 2 );
                            tc_.movePosition ( QTextCursor::Left,
                                               QTextCursor::KeepAnchor, 1 );
                            if ( tc_.selectedText () == "\"" ) {
                                tc_.clearSelection ();
                                tc_ = textCursor ();
                                tc_.insertText ( "\"\"" );
                                tc_.movePosition ( QTextCursor::Left,
                                                   QTextCursor::MoveAnchor, 1 );
                                setTextCursor ( tc_ );
                            }
                        }
                        break;
                }
                ensureCursorVisible ();
                break;
            case Qt::GroupSwitchModifier:
            // keyboard selection
            case Qt::ShiftModifier:
                if ( e->key () >= Qt::Key_Space && e->key ()
                        <= Qt::Key_AsciiTilde && e->key () != Qt::Key_ParenLeft
                        && e->key () != Qt::Key_ParenRight && e->key ()
                        != Qt::Key_QuoteDbl && e->key () ) {
                    checkWritingPosition ();
                    QTextEdit::keyPressEvent ( e );
                    ensureCursorVisible ();
                    break;
                }

                switch ( e->key () ) {
                    // select from current position to prompt
                    case Qt::Key_Home:
                        actionHome ( true );
                        break;

                    // select from current position end of line
                    case Qt::Key_End:
                        actionEnd ( true );
                        break;

                    // select one char (left or right of cursor)
                    case Qt::Key_Left:
                    case Qt::Key_Right:
                        if ( textCursor ().position () > lineStart_ ) {
                            QTextEdit::keyPressEvent ( e );
                        }
                        break;

                    // insert closing parenthesis automatically
                    case Qt::Key_ParenLeft:
                        QTextEdit::keyPressEvent ( e );
                        tc_ = textCursor ();
                        tc_.insertText ( ")" );
                        tc_.movePosition ( QTextCursor::Left,
                                           QTextCursor::MoveAnchor, 1 );
                        setTextCursor ( tc_ );
                        break;

                    // step over closing parenthesis
                    case Qt::Key_ParenRight:
                        tc_ = textCursor ();
                        tc_.movePosition ( QTextCursor::Right,
                                           QTextCursor::KeepAnchor, 1 );
                        if ( tc_.selectedText () != ")" ) {
                            QTextEdit::keyPressEvent ( e );
                        } else {
                            tc_ = textCursor ();
                            tc_.movePosition ( QTextCursor::Right,
                                               QTextCursor::MoveAnchor, 1 );
                            setTextCursor ( tc_ );
                        }
                        break;

                    // insert closing quotes automatically
                    case Qt::Key_QuoteDbl:
                        tc_ = textCursor ();
                        tc_.movePosition ( QTextCursor::Right,
                                           QTextCursor::KeepAnchor, 1 );
                        if ( tc_.selectedText () != "\"" ) {
                            QTextEdit::keyPressEvent ( e );
                            QTextEdit::keyPressEvent ( e );
                            tc_ = textCursor ();
                            tc_.movePosition ( QTextCursor::Left,
                                               QTextCursor::MoveAnchor, 1 );
                        } else {
                            tc_ = textCursor ();
                            tc_.movePosition ( QTextCursor::Right,
                                               QTextCursor::MoveAnchor, 1 );
                        }
                        setTextCursor ( tc_ );
                        break;
                }
                ensureCursorVisible ();
                break;
        }
    }
}




/**
 * \brief Print a string on the console.
 *
 * \param print String to print.
 */
void UCQConsole::print ( QString print )
{
    append ( print );
}

/**
 * \brief Prints the character of the letter.
 *
 * \param l Letter which to print.
 */
void UCQConsole::printLetter ( Letter * l )
{
    if ( executing_ )
        return;

    if ( l != 0 ) {
        QString t ( "" );
        for ( int i = 0; i < l->getCodepoint ().size (); ++i ) {
            t.append(UCTools::unicodeToString(l->getCodepoint ().at ( i )));
        }
        checkWritingPosition ();
        insertPlainText ( t );
    }
}

/**
 * \brief Prints the linestring of the items that are selected in the TableView
 *        on the console.
 *
 * \param selection Selected Letters in the TableView
 */
void UCQConsole::printSelection ( QList<Letter*> selection )
{
    if ( executing_ )
        return;

    if ( !selection.isEmpty () ) {
        for ( int i = 0; i < selection.size (); ++i )
        {
            print ( selection.at ( i )->getLinestring () );
        }
        print ( "" );
        displayPrompt ();
    }
}

/**
 * \brief Sets the font of the console.
 *
 * \param font The font that will be used after the call.
 */
void UCQConsole::setFont(const QFont& font){
    QFont f = font;
    selectAll();
    setCurrentFont(f);
    tc_ = textCursor();
    tc_.clearSelection();
    setTextCursor(tc_);
}

/**
 * \brief Setter for the history.
 * \param h History.
 */
void UCQConsole::setHistory ( const QStringList &h )
{
    history_ = h;
    historyCount_ = history_.size ();
    history_.append ( "" );
}

/**
 * \brief Sets the Thread that will be used for executing the commands.
 *
 * \param lua Thread that will be used for executing the commands.
 */
void UCQConsole::setExecuteThread ( UCQExecuteLuaThread * executeThread )
{
    // connect new controller
	executeLuaThread_ = executeThread;
}

/**
 * \brief Sets the prompt.
 *
 * \param p Prompt.
 */
void UCQConsole::setPrompt ( const QString& p )
{
    prompt_ = p;
}

/**
 * \brief Extends the standard context menu with a "Settings"-action.
 *
 * \param event The event that triggers the context menu.
 */
void UCQConsole::contextMenuEvent(QContextMenuEvent *event)
{
    QMenu *menu = createStandardContextMenu();
    QAction *actionSettings;
    actionSettings = new QAction(menu);
    actionSettings->setText(tr("Console &Settings"));
    menu->addAction(actionSettings);
    connect(actionSettings, SIGNAL(triggered()), this, SLOT(showSettings()));

    menu->exec(event->globalPos());
    delete menu;
}

/**
 * \brief Shows a font dialog to choose the font for the console.
 */
void UCQConsole::showSettings(){
    bool ok;
    QFont font = QFontDialog::getFont( &ok, currentFont(), this );
    if (ok) {
        setFont(font);
    }
}

void UCQConsole::stopExec(){
    emit stopExecution ();
    executeLuaThread_->exit(0);
}
